import re
from abc import ABC, abstractmethod
from datetime import datetime
from typing import Type, Tuple

DEFAULT_DATE_FORMAT = "%Y-%m-%d %H:%M:%S"


class Validator(ABC):

    @staticmethod
    def assert_valid(errors: list, validation: Tuple[bool, str], throw_if_invalid: bool = False) -> None:
        if not validation[0]:
            if throw_if_invalid:
                raise AssertionError(validation[1])
            else:
                errors.append(validation[1])

    @staticmethod
    def is_not_blank(input_string: str, min_length: int = 0) -> bool:
        return input_string and len(input_string) >= min_length

    @staticmethod
    def matches(
            input_string: str,
            regex_pattern: str) -> bool:

        return bool(re.match(regex_pattern, input_string))

    @staticmethod
    def is_integer(
            number: str,
            min_value: int = 0,
            max_value: int = 65535) -> bool:

        return number.isdigit() and min_value <= int(number) <= max_value

    @staticmethod
    def is_float(
            number: str,
            min_value: float = 0,
            max_value: float = 65535) -> bool:

        return number.isdecimal() and min_value <= float(number) <= max_value

    @staticmethod
    def is_enum(
            name: str,
            enum_type: Type) -> bool:

        return name.upper() in enum_type.__dict__

    @staticmethod
    def is_date(
            date_text: str,
            fmt: str = DEFAULT_DATE_FORMAT) -> bool:

        try:
            datetime.strptime(date_text, fmt)
            return True
        except ValueError:
            return False

    @staticmethod
    def has_no_nulls(*args):
        return all(opt is not None for opt in args)

    @abstractmethod
    def __call__(self, *args, **kwargs) -> bool:
        pass



import os
import re
import sys
import termios
import tty
from typing import Optional, Tuple, List

assert sys.stdin.isatty(), 'This module requires a terminal (TTY)'


def screen_size() -> Optional[List[str]]:
    """Retrieve the size of the terminal"""
    if sys.stdout.isatty():
        return os.popen('stty size').read().split()
    else:
        return None


def get_cursor_position() -> Optional[Tuple[int, int]]:
    """ Get the terminal cursor position
    Solution taken from:
    - https://stackoverflow.com/questions/46651602/determine-the-terminal-cursor-position-with-an-ansi-sequence-in-python-3
    """
    if sys.stdout.isatty():
        buf = ""
        stdin = sys.stdin.fileno()
        attrs = termios.tcgetattr(stdin)
        try:
            tty.setcbreak(stdin, termios.TCSANOW)
            sys.stdout.write("\x1b[6n")
            sys.stdout.flush()
            while True:
                buf += sys.stdin.read(1)
                if buf[-1] == "R":
                    break
        finally:
            termios.tcsetattr(stdin, termios.TCSANOW, attrs)
        try:
            matches = re.match(r"^\x1b\[(\d*);(\d*)R", buf)
            groups = matches.groups()
        except AttributeError:
            return None
        return int(groups[0]), int(groups[1])
    else:
        return None


def set_enable_echo(enable: bool = True) -> None:
    """Enable echo in the terminal
    :param enable:
    """
    if sys.stdout.isatty():
        os.popen(f"stty {'echo -raw' if enable else 'raw -echo min 0'}").read()



import os
import sys

from hspylib.core.tools.commons import __version__, __curdir__, sysout, __here__
from hspylib.modules.application.application import Application

HERE = __here__(__file__)


class Main(Application):
    """TODO"""

    # The application name
    APP_NAME = os.path.basename(__file__)

    # The application version
    VERSION = __version__(f"{HERE}/.version")

    # Usage message
    USAGE = (HERE / "usage.txt").read_text().format('.'.join(map(str, VERSION)))

    def __init__(self, app_name: str):
        super().__init__(app_name, self.VERSION, self.USAGE, __curdir__(__file__))

    def main(self, *params, **kwargs) -> None:
        """Run the application with the command line arguments"""
        sysout(f'Hello {self.APP_NAME}')


if __name__ == "__main__":
    """Application entry point"""
    Main('Application name').INSTANCE.run(sys.argv[1:])



import re
from typing import Any


class Argument:

    def __init__(self, arg_name: str, validation_regex: str = '.*', required: bool = True, next_in_chain: Any = None):
        self.arg_name = arg_name
        self.validation_regex = validation_regex
        self.required = required
        self.next_in_chain = next_in_chain
        self.value = ''

    def __str__(self):
        return "arg_nam: {}, validation_regex: {}, required: {}, value: {}, next: {}" \
            .format(self.arg_name, self.validation_regex, self.required, self.value, self.next_in_chain)

    def __repr__(self):
        return str(self)

    def set_value(self, provided_arg: str) -> bool:
        self.value = provided_arg if re.match(rf'^({self.validation_regex})$', provided_arg) else None
        return True if self.value else False

    def set_next(self, argument: Any) -> None:
        self.next_in_chain = argument



import re
from abc import ABC


class Vt100(ABC):
    """
    References:
        - https://vt100.net/docs/vt100-ug/chapter3.html
        - https://espterm.github.io/docs/VT100%20escape%20codes.html
    """

    # Esc<Sequence>
    @classmethod
    def escape(cls, seq: str) -> str:
        return f"\033{seq}"

    # Esc[<Code>
    @classmethod
    def sequence(cls, code: str) -> str:
        return cls.escape(f"[{code}")

    # Esc7
    @classmethod
    def save_cursor(cls):
        return cls.escape('7')

    # Esc8
    @classmethod
    def restore_cursor(cls):
        return cls.escape('8')

    # Esc[c
    @classmethod
    def reset(cls):
        return cls.sequence('c')

    # Esc[?7<h/l>
    @classmethod
    def set_auto_wrap(cls, enabled: bool) -> str:
        return cls.sequence(f"?7{'h' if enabled else 'l'}")

    # Esc[?25<h/l>
    @classmethod
    def set_show_cursor(cls, enabled: bool) -> str:
        return cls.sequence(f"?25{'h' if enabled else 'l'}")

    # Esc[<Modes...>m
    @classmethod
    def mode(cls, mod_seq: str) -> str:
        assert re.match(r"[0-9]+(;[0-9]+){0,2}", mod_seq)
        return cls.sequence(f"{mod_seq}m")

    # Esc[<n>J
    @classmethod
    def clear_screen(cls, mod_cls: int = None) -> str:
        if mod_cls is None:
            return cls.sequence('J')
        else:
            assert mod_cls in [0, 1, 2]
            return cls.sequence(f'{mod_cls}J')

    # Esc[<n>K
    @classmethod
    def clear_line(cls, mod_cls: int = None) -> str:
        if mod_cls is None:
            return cls.sequence('K')
        else:
            assert mod_cls in [0, 1, 2]
            return cls.sequence(f'{mod_cls}K')

    # Esc[<v>;<h>H
    @classmethod
    def cursor_pos(cls, cup_seq: str = None) -> str:
        if cup_seq is None:
            return cls.sequence('H')
        else:
            assert re.match(r"[0-9]*;[0-9]*", cup_seq)
            return cls.sequence(f"{cup_seq}H")

    # Esc[<n><A/B/C/D>
    @classmethod
    def cursor_move(cls, amount: int, direction: str) -> str:
        assert int(amount) >= 0 and direction in ['A', 'B', 'C', 'D']
        return cls.sequence(f"{amount}{direction}")

    # Esc[<n>A
    @classmethod
    def cursor_move_up(cls, amount: int = None) -> str:
        return cls.cursor_move(amount if amount else 0, 'A')

    # Esc[<n>B
    @classmethod
    def cursor_move_down(cls, amount: int = None) -> str:
        return cls.cursor_move(amount if amount else 0, 'B')

    # Esc[<n>C
    @classmethod
    def cursor_move_forward(cls, amount: int = None) -> str:
        return cls.cursor_move(amount if amount else 0, 'C')

    # Esc[<n>D
    @classmethod
    def cursor_move_backward(cls, amount: int = None) -> str:
        return cls.cursor_move(amount if amount else 0, 'D')



import logging as log


class HSBaseException(Exception):
    """This is a generic exception and should not be raised. It may be inherited instead"""
    def __init__(self, message: str, cause: Exception = None):
        fmt_msg = f'{message}' + (' => {str(cause)}' if cause else '')
        super().__init__(fmt_msg)
        log.error(fmt_msg)


class EntityNotFoundError(HSBaseException):
    """Raised when an entity is not found"""


class InputAbortedError(HSBaseException):
    """Raised when an input method is aborted"""


class ProgrammingError(HSBaseException):
    """Exception raised for programming errors, e.g. table not found
    or already exists, syntax error in the SQL statement, wrong number
    of parameters specified, etc."""


class InvalidArgumentError(HSBaseException):
    """Raised when an invalid argument is received by the application"""


class InvalidOptionError(HSBaseException):
    """Raised when an invalid option is received by the application"""



import getopt
import logging as log
import signal
import sys
from datetime import datetime
from typing import List, Callable, Optional, Tuple, Set

from hspylib.core.config.app_config import AppConfigs
from hspylib.core.exception.exceptions import InvalidOptionError, InvalidArgumentError
from hspylib.core.meta.singleton import Singleton
from hspylib.core.tools.commons import sysout, syserr
from hspylib.modules.application.argument import Argument
from hspylib.modules.application.argument_chain import ArgumentChain
from hspylib.modules.application.option import Option


class Application(metaclass=Singleton):
    """HSPyLib application framework"""

    def __init__(
            self,
            app_name: str,
            app_version: Tuple[int, int, int],
            app_usage: str = "Usage: __main__.py [options] <arguments>",
            source_dir: str = None,
            resource_dir: str = None,
            log_dir: str = None):

        signal.signal(signal.SIGINT, self.exit_handler)
        self.app_name = app_name
        self.app_version = app_version
        self.app_usage = app_usage
        self.options = {}
        self.cond_args_chain = {}
        self.args = None
        if source_dir:
            self.configs = AppConfigs(
                source_root=source_dir,
                resource_dir=resource_dir,
                log_dir=log_dir
            )
        self.with_option('h', 'help', handler=self.usage)
        self.with_option('v', 'version', handler=self.version)

    def main(self, *args, **kwargs):
        """Execute the application's main statements"""
        log.info('Main entry point handler called. nothing to do')

    def run(self, *args, **kwargs):
        """Main entry point handler"""
        log.info('Run started {}'.format(datetime.now()))
        try:
            self.main(*args, **kwargs)
            self.exit_handler()
        except (InvalidArgumentError, InvalidOptionError) as err:
            syserr(str(err))
            self.usage(1)
        finally:
            log.info('Run finished {}'.format(datetime.now()))

    def cleanup(self):
        """TODO"""
        log.info('Cleanup handler called. nothing to cleanup')

    def exit_handler(self, signum=0, frame=None, clear_screen: bool = False) -> None:
        """
        Handle interruptions to shutdown gracefully
        :param signum: The signal number or the exit code
        :param frame: The frame raised by the signal
        :param clear_screen: Whether to clean the screen before execution or not
        """
        if frame is not None:
            log.warn('Signal handler hooked signum={} frame={}'.format(signum, frame))
            exit_code = 3
        else:
            log.info('Exit handler called')
            exit_code = signum
        self.cleanup()
        if clear_screen:
            sysout('%ED2%%HOM%')
        sys.exit(exit_code)

    def usage(self, *args) -> None:
        """Display the usage message and exit with the specified code ( or zero as default )
        :param args: The application arguments
        """
        sysout(self.app_usage)
        self.exit_handler(args[0] or 0)

    def version(self, *args) -> None:
        """Display the current program version and exit
        :param args: The application arguments
        """
        sysout('{} v{}'.format(self.app_name, '.'.join(map(str, self.app_version))))
        self.exit_handler(args[0] or 0)

    def parse_parameters(self, parameters: List[str]) -> None:
        """ Handle program parameters.
        :param parameters: The list of unparsed program parameters passed by the command line
        """
        # First parse all options, then, parse arguments
        self.parse_arguments(self.parse_options(parameters))

    def parse_options(self, parameters: List[str]) -> List[str]:
        """Handle program options in the form: Short opts: -<C>, Long opts: --<Word>
        :param parameters: The list of unparsed program parameters passed by the command line
        :return A list of remaining arguments
        """
        try:
            opts, args = getopt.getopt(parameters, self._shortopts(), self._longopts())
            for op, arg in opts:
                opt = self._getopt(op)
                if opt and opt.cb_handler:
                    opt.cb_handler(arg)
                else:
                    raise InvalidOptionError(f'Option "{op}" not recognized')
            return args
        except getopt.GetoptError as err:
            raise InvalidOptionError(f"### Invalid option: {str(err)}")

    def parse_arguments(self, provided_args: List[str]) -> None:
        """Handle program arguments
        :param provided_args: Arguments left after processing all options
        """
        valid = False
        if self.cond_args_chain:
            # Walk through all argument conditions in chain and validate the provided arguments
            for cond_arg in self.cond_args_chain:
                arg = cond_arg.argument
                try:
                    self.recursive_set(0, arg, provided_args)
                    # At this point, we found a valid argument chain
                    valid = self.validate_args_in_chain(arg)
                    # Will be true if all required args were set
                    self.args = provided_args
                    break
                except getopt.GetoptError as err:
                    log.debug(str(err))
                    continue  # Just try the next chain
                except LookupError:
                    continue  # Just try the next chain
            if not valid:
                raise InvalidArgumentError(f"Invalid arguments  \"{', '.join(provided_args)}\"")

    def recursive_set(self, idx: int, argument: Argument, provided_args: List[str]):
        """ Try to set a value for each provided argument. If any failure setting occur, raise an exception to be caught
            by the parse function, so they can try the next chain in the loop.
        """
        if not argument or idx >= len(provided_args):
            return
        if not argument.set_value(provided_args[idx]):
            raise LookupError(
                f'Invalid argument "{provided_args[idx]}"')
        else:
            self.recursive_set(idx + 1, argument.next_in_chain, provided_args)

    @staticmethod
    def validate_args_in_chain(arg: Argument) -> bool:
        missing = 0
        next_arg = arg
        while next_arg:
            missing += 1 if not next_arg.value and next_arg.required else 0
            next_arg = next_arg.next_in_chain
        return missing == 0

    def with_option(
            self,
            shortopt: chr,
            longopt: str,
            has_argument: bool = False,
            handler: Callable = None):
        self.options[longopt] = Option(shortopt, longopt, has_argument, handler)

    def with_arguments(
            self,
            chained_args: Set[ArgumentChain.ChainedArgument]):
        self.cond_args_chain = chained_args

    def _shortopts(self) -> str:
        return ''.join([op.shortopt.replace('-', '') for key, op in self.options.items()])

    def _longopts(self) -> List[str]:
        return [op.longopt.replace('-', '') + ' ' for key, op in self.options.items()]

    def _getopt(self, item: str) -> Optional[Option]:
        return next((op for key, op in self.options.items() if op.is_eq(item)), None)
